/* --------------------------------------------------------

   Project:
      External_Ballistics

    Purpose:
      Simulates the flight of a round ball

   Author:
      Stephen C. Wardlaw, M.D.
      IVDx Consulting
      Highrock
      Lyme, CT 06371
      <swardlaw@ivdx.com>

   Edit History:
      30 March, 2011

 -------------------------------------------------------- */
#include "external_ballistics.h"
#include <QtCore>
#include <math.h>

const double grains_per_pound = 7000;// Grains per pound
const double dens_lead = 0.4093;	// Density of lead in lbs/cu.in.
const double _pi_ = 3.14159;
const double g_English = 32.174;	// g - in English units ft/sec/sec
const double mph_to_fps	= 1.46667;	// MPH to FPS conversion
const double in_per_foot = 12.0;		// Inches per foot
const double drag_coeff	= 0.0755;	// Arbitrary drag coeff. This makes the British Retardation data match real-world data. DON'T CHANGE!
const double BC = 1.0;	// Relative ballistic coefficient. Tweak this if you must, not 'drag_coeff'
const int data_interval = 1;	// Data every x yards
const double std_temp_f = 70.0;	// 'Standard' sea-level temperature
const double temp_lapse_rate = 3.55E-3;	// Environmental lapse rate in deg F/ft
const int d_spc = 10;		// Output line data spacing


// --------------------------------------------------------
//			CONSTRUCTOR AND DESTRUCTOR
// --------------------------------------------------------
External_Ballistics::External_Ballistics(QWidget *parent, Qt::WFlags flags)
	: QMainWindow(parent, flags)
{
	ui.setupUi(this);

	connect(ui.mEditCaliber,SIGNAL(editingFinished()),this,SLOT(caliberChanged()));
	connect(ui.mEditWeight,SIGNAL(editingFinished()),this,SLOT(weightChanged()));
	connect(ui.mEditAim,SIGNAL(editingFinished()),this,SLOT(aimChanged()));
	connect(ui.mEditVelocity,SIGNAL(editingFinished()),this,SLOT(mvChanged()));
	connect(ui.mEditSight,SIGNAL(editingFinished()),this,SLOT(sightHeightChanged()));
	connect(ui.mEditWind,SIGNAL(editingFinished()),this,SLOT(windChanged()));
	connect(ui.mEditTemp,SIGNAL(editingFinished()),this,SLOT(tempChanged()));
	connect(ui.mRunSimulation,SIGNAL(clicked()),this,SLOT(fireClicked()));
	connect(ui.mActShowForce,SIGNAL(activated()),this,SLOT(createPressureCurve()));

	mCaliber = 0.50;
	mWeight = calcBallWeight(mCaliber);
	mAimPoint = 50.0;
	mSightHeight = 1.0;
	mXWind = 0.0;
	mMuzzleVelocity = 1400.0;
	mTemp = 70.0;	// Seventy def F as std temp

	// Set up the altitude selection
	ui.mAltitude->addItem("Sea Level",QVariant(0));
	double alt = 1000.0;
	while(alt < 10000.0) {
		ui.mAltitude->addItem(QString("%1").arg(alt,0,'f',0),alt);
		alt += 1000.0;
	}
	while(alt <= 16000.0) {
		ui.mAltitude->addItem(QString("%1").arg(alt,0,'f',0),alt);
		alt += 2000.0;
	}
	// Not, connect it...
	connect(ui.mAltitude,SIGNAL(currentIndexChanged(int)),this,SLOT(altitudeChanged(int)));

	showInput();

	ui.mEditCaliber->setModified(false);

	calcRelDensity();
}

External_Ballistics::~External_Ballistics()
{

}
// --------------------------------------------------------
//			SLOTS
// --------------------------------------------------------
void
External_Ballistics::caliberChanged()
{
	bool ok;

	if(ui.mEditCaliber->isModified()) {
		double cal = ui.mEditCaliber->text().toDouble(&ok);
		if(!ok) {
			showLine(QString("<b>Error</b> Caliber is not a valid number"));
			showInput();
			return;
		}
		if(cal < 0.05) {
			showLine(QString("<b>Error</b> Caliber must be greater than 0.05"));
			showInput();
			return;
		}
		if(cal > 1.5) {
			showLine(QString("<b>Error</b> Caliber must be less than 1.5"));
			showInput();
			return;
		}
		mCaliber = cal;
		mWeight = calcBallWeight(mCaliber);
	}
	showInput();
}
// --------------------------------------------------------
void
External_Ballistics::aimChanged()
{
	bool ok;
	double val;

	if(ui.mEditAim->isModified()) {
		val = ui.mEditAim->text().toDouble(&ok);
		if(!ok) {
			showLine(QString("<b>Error</b> Aim point is not a valid number"));
			showInput();
			return;
		}
		if(val < 5.0) {
			showLine(QString("<b>Error</b> Aim point must be greater than 5 yds."));
			showInput();
			return;
		}
		if(val > 200) {
			showLine(QString("<b>Error</b> Aim point must be less than 200 yds."));
			showInput();
			return;
		}
		mAimPoint = val;
	}
	showInput();
}
// --------------------------------------------------------
void
External_Ballistics::altitudeChanged(int index)
{
	double alt = ui.mAltitude->itemData(index).toDouble();
	mTemp = lapsedTemp(alt);
	calcRelDensity();
	showInput();
}
// --------------------------------------------------------------------
// Make a file showing the force on the projectile vs velocity
void
External_Ballistics::createPressureCurve()
{
	QFile file(QString("forces.csv"));
	if(file.open(QIODevice::WriteOnly)) {
		QTextStream stream(&file);
		double bulletArea = mCaliber*mCaliber*_pi_/4.0;
		stream << "Velocity" << "\tForce" << endl;
		double vel = 100.0;	// Begin at 100 fps
		while(vel <= 3000.0) {
			double force = rValue(vel)*bulletArea*drag_coeff*mRelDensity/BC;
			stream << QString("%1\t%2").arg(vel,0,'f',0).arg(force,0,'f',2) << endl;
			vel += 50.0;
		}
	}
	ui.mShowResults->clear();
	showLine("Force file created as 'forces.csv'");

}
// --------------------------------------------------------
void
External_Ballistics::mvChanged()
{
	bool ok;
	double val;

	if(ui.mEditVelocity->isModified()) {
		val = ui.mEditVelocity->text().toDouble(&ok);
		if(!ok) {
			showLine(QString("<b>Error</b> MV is not a valid number"));
			showInput();
			return;
		}
		if(val < 100.0) {
			showLine(QString("<b>Error</b> MV must be greater than 100 fps"));
			showInput();
			return;
		}
		if(val > 4000.0) {
			showLine(QString("<b>Error</b> MV must be less than 4000 fps"));
			showInput();
			return;
		}
		mMuzzleVelocity = val;
	}
	showInput();
}
// --------------------------------------------------------
void
External_Ballistics::sightHeightChanged()
{
	bool ok;
	double val;

	if(ui.mEditSight->isModified()) {
		val = ui.mEditSight->text().toDouble(&ok);
		if(!ok) {
			showLine(QString("<b>Error</b> Sight height is not a valid number"));
			showInput();
			return;
		}
		if(val < 0.0) {
			showLine(QString("<b>Error</b> Sight height must be greater than 0 in."));
			showInput();
			return;
		}
		if(val > 12.0) {
			showLine(QString("<b>Error</b> Sight height must be less than 12 in."));
			showInput();
			return;
		}
		mSightHeight = val;
	}
	showInput();
}
// --------------------------------------------------------
void
External_Ballistics::tempChanged()
{
	bool ok;
	double val;

	if(ui.mEditTemp->isModified()) {
		val = ui.mEditTemp->text().toDouble(&ok);
		if(!ok) {
			showLine(QString("<b>Error</b> Temperature is not a valid number"));
			showInput();
			return;
		}
		if(val < -10.0) {
			showLine(QString("<b>Error</b> Temp. must be greater than -10 deg."));
			showInput();
			return;
		}
		if(val > 120.0) {
			showLine(QString("<b>Error</b> Temp. must be less than 120 deg."));
			showInput();
			return;
		}
		mTemp = val;
	}
	calcRelDensity();
	showInput();
}
// --------------------------------------------------------
void
External_Ballistics::weightChanged()
{
	bool ok;
	double val;

	if(ui.mEditWeight->isModified()) {
		val = ui.mEditWeight->text().toDouble(&ok);
		if(!ok) {
			showLine(QString("<b>Error</b> Weight is not a valid number"));
			showInput();
			return;
		}
		if(val < 1.0) {
			showLine(QString("<b>Error</b> Weight must be greater than 1 gr."));
			showInput();
			return;
		}
		if(val > 7000.0) {
			showLine(QString("<b>Error</b> Weight must be less than 7000 gr."));
			showInput();
			return;
		}
		mWeight = val;
	}
	showInput();
}
// --------------------------------------------------------
void
External_Ballistics::windChanged()
{
	bool ok;
	double val;

	if(ui.mEditWind->isModified()) {
		val = ui.mEditWind->text().toDouble(&ok);
		if(!ok) {
			showLine(QString("<b>Error</b> Wind is not a valid number"));
			showInput();
			return;
		}
		if(val < 0.0) {
			showLine(QString("<b>Error</b> Wind must be greater than 0"));
			showInput();
			return;
		}
		if(val > 60.0) {
			showLine(QString("<b>Error</b> Wind must be less than 60 mph"));
			showInput();
			return;
		}
		mXWind = val;
	}
	showInput();
}
// --------------------------------------------------------
void
External_Ballistics::fireClicked()
{
	ui.mShowResults->clear();
	calculate();
	showLine(QString("%1 %2 %3 %4 %5 %6").arg("Yards",d_spc).arg("Vel",d_spc).arg("Fpe",d_spc).arg("Drop",d_spc).arg("Drift",d_spc).arg("TOF",d_spc));
	showOutput();

	// See if we need to make a csv file
	if(ui.mCreateFile->isChecked()) {
		QFile file(QString("%1.csv").arg(makeFileName()));
		if(file.open(QIODevice::WriteOnly)) {
			QTextStream stream(&file);
			stream << "Yards" << "\tVelocity" << "\tFpe" << "\tDrop" << "\tDrift" << "\tTOF" << endl;

			int yards = 0;
			double vel;
			double energy;
			double drop;

			// Begin with data every yard
			while((yards < mVelocity.count()) && (yards * data_interval) < 50) {
				vel = mVelocity.at(yards);
				energy = vel*vel*mWeight/(2.0*g_English*grains_per_pound);
				drop = mSlope*yards - (mDrop.at(yards) + mSightHeight);
				stream << QString("%1\t%2\t%3\t%4\t%5\t%6").arg(yards).arg(vel,0,'f',0).arg(energy,0,'f',1).arg(drop,0,'f',1).arg(mDrift.at(yards),0,'f',1).arg(mTof.at(yards),0,'f',3);
				stream << endl;
				++yards;
			}

			// At 50-100 yards, go up by 2
			while((yards < mVelocity.count()) && (yards * data_interval) < 100) {
				vel = mVelocity.at(yards);
				energy = vel*vel*mWeight/(2.0*g_English*grains_per_pound);
				drop = mSlope*yards - (mDrop.at(yards) + mSightHeight);
				stream << QString("%1\t%2\t%3\t%4\t%5\t%6").arg(yards).arg(vel,0,'f',0).arg(energy,0,'f',1).arg(drop,0,'f',1).arg(mDrift.at(yards),0,'f',1).arg(mTof.at(yards),0,'f',3);
				stream << endl;
				yards += 2;
			}

			// Now go up by 5 yard intervals
			while((yards < mVelocity.count()) && (yards * data_interval) < 400) {
				vel = mVelocity.at(yards);
				energy = vel*vel*mWeight/(2.0*g_English*grains_per_pound);
				drop = mSlope*yards - (mDrop.at(yards) + mSightHeight);
				stream << QString("%1\t%2\t%3\t%4\t%5\t%6").arg(yards).arg(vel,0,'f',0).arg(energy,0,'f',1).arg(drop,0,'f',1).arg(mDrift.at(yards),0,'f',1).arg(mTof.at(yards),0,'f',3);
				stream << endl;
				yards += 5;
			}

		}
	}
}
// --------------------------------------------------------
//			PRIVATE METHODS
// --------------------------------------------------------
double
External_Ballistics::calcBallWeight(double cal)
{
	cal /= 2.0;	// Get the radius
	double vol = cal*cal*cal*_pi_*4.0/3.0;
	return vol*dens_lead*grains_per_pound;
}
// --------------------------------------------------------------------
void
External_Ballistics::calcRelDensity()
{
	const double std_pres = 14.7;	// Standard air pressure at sea level

	// Calculate the air pressure at the altitude
	double alt = ui.mAltitude->itemData(ui.mAltitude->currentIndex()).toDouble();
	double pres = 14.7*pow((1 - 6.8877E-6*alt),5.25588);	// Pressure at altitude in ft.

	// Adjust pressure for non-standard temp
	const double abs_f = 459.7;	// 
	double relTP = (abs_f + lapsedTemp(alt))/(abs_f + mTemp);
	pres *= relTP;	// Pressure adjusted for temperature
	mRelDensity = pres/std_pres;
}
// --------------------------------------------------------------------
// Performs the calculation
void
External_Ballistics::calculate()
{
	const int max_intervals = 2000;
	const double time_interval = 0.001;	// Integration interval
	
	double hVel = mMuzzleVelocity;	// Horizontal velocity
	double hPosn = 0;	// Horizontal distance
	double vPosn = 0;	// Vertical distance
	double xVel = 0;		// Cross-velocity
	double xPosn = 0;	// Cross-position
	double currTime = 0;	
	double bulletMass = mWeight/(grains_per_pound*g_English);
	double bulletArea = mCaliber*mCaliber*_pi_/4.0;
	mSlope = 0.0;
		
	mVelocity.clear();
	mDrop.clear();
	mDrift.clear();
	mTof.clear();

	int intervalNum = 0;
	int yardNum = 0;
	
	while(intervalNum < max_intervals) {
		// To
		currTime = intervalNum*time_interval;

		// Increment distance traveled over interval
		hPosn += hVel*time_interval;
		xPosn += xVel*time_interval;
		
		// Vertical fall due to gravity
		// s = 0.5 at^2
		vPosn = 0.5*g_English*currTime*currTime;
		
		// Find force on bullet
		// Total velocity is the vector sum of the velocity and crosswind
		double xFPS = mXWind*mph_to_fps - xVel;
		double vT = ::sqrt(hVel*hVel + xFPS*xFPS);
		double force = rValue(vT)*bulletArea*drag_coeff*mRelDensity/BC;

		// Resolve force into cross and forward vectors
		double xForce = force*xFPS/hVel;
		double hForce = force*(hVel - xFPS)/hVel;

		double hAccel = hForce/bulletMass;
		double xAccel = xForce/bulletMass;
		double deltaHV = hAccel*time_interval;
		double deltaXV = xAccel*time_interval;
		hVel -= deltaHV;
		xVel += deltaXV;

		// See if we are at a yard marker - we take data every yard...
		if((hPosn/3.0) > (double)(yardNum*data_interval)) {
			mVelocity.append(vT);
			mDrift.append(xPosn*in_per_foot);
			mDrop.append(vPosn*in_per_foot);
			mTof.append(currTime);	// TOF in seconds
			++yardNum;
		}

		++intervalNum;
	}
/* 
	Calculate the angle of the barrel based upon the drop of the
 	projectule at the aim point. The total drop is the ballistic
 	drop plus the sight height
*/
	// The slope is measured in inches/yard
	int posn = (int)mAimPoint/data_interval;
	if(mSlope == 0.0) {
		mSlope = (mDrop.at(posn) + mSightHeight)/mAimPoint;
	}

}
// --------------------------------------------------------------------
void
External_Ballistics::doOutputLine(int yards)
{
	double vel = mVelocity.at(yards);
	if(vel > 0.0) {
		double energy = vel*vel*mWeight/(2.0*g_English*grains_per_pound);
		showLine(QString("%1 %2 %3 %4 %5 %6").arg(yards,d_spc).arg(vel,d_spc,'f',0).arg(energy,d_spc,'f',0).arg(mSlope*yards - (mDrop.at(yards) + mSightHeight),d_spc,'f',1).arg(mDrift.at(yards),d_spc,'f',1).arg(mTof.at(yards),d_spc,'f',3));
	}
}
// --------------------------------------------------------------------
// 'Standard' temperature at a given altitude
double
External_Ballistics::lapsedTemp(double alt)
{
	return std_temp_f - temp_lapse_rate*alt;
}
// --------------------------------------------------------------------
// Make up a file name from the data
QString
External_Ballistics::makeFileName()
{
	QString name = QString("c%1_w%2_v%3_x%4").arg(mCaliber,0,'f',2).arg(mWeight,0,'f',1).arg(mMuzzleVelocity,0,'f',0).arg(mXWind,0,'f',0);
	name = name.replace('.','d');	// Don't allow periods...
	return name;
}
// --------------------------------------------------------------------
// Calculates retardation as a function of velocity, based upon the 1904 British
// data.
double
External_Ballistics::rValue(double vel)
{
	// British data
	const double fact_m1	 	= 74422/1.0E8;
	const double fact_m2 	= 59939/1.0E12;
	const double fact_m3 	= 23385/1.0E22;
	const double fact_m4 	= 95408/1.0E12;
	const double fact_m5 	= 59814/1.0E8;
	const double fact_m6 	= 58497/1.0E7;
	const double fact_m7 	= 15366/1.0E7;
	const double expn_1	= 1.6;
	const double expn_2	= 3.0;
	const double expn_3	= 6.45;
	const double expn_4	= 3.0;
	const double expn_5	= 1.8;
	const double expn_6	= 1.5;
	const double expn_7	= 1.67;
	
	if(vel <= 840) {
		return ::pow(vel,expn_1)*fact_m1;
	} else if(vel <= 1040) {
		return ::pow(vel,expn_2)*fact_m2;
	} else if(vel <= 1190) {
		return ::pow(vel,expn_3)*fact_m3;
	} else if(vel <= 1460) {
		return ::pow(vel,expn_4)*fact_m4;
	} else if(vel <= 2000) {
		return ::pow(vel,expn_5)*fact_m5;
	} else if(vel <= 2600) {
		return ::pow(vel,expn_6)*fact_m6;
	} else {
		return ::pow(vel,expn_7)*fact_m7;
	}
}
// --------------------------------------------------------
void
External_Ballistics::showInput()
{
	ui.mEditCaliber->setText(QString("%1").arg(mCaliber,0,'f',2));
	ui.mEditWeight->setText(QString("%1").arg(mWeight,0,'f',1));
	ui.mEditSight->setText(QString("%1").arg(mSightHeight,0,'f',1));
	ui.mEditVelocity->setText(QString("%1").arg(mMuzzleVelocity,0,'f',0));
	ui.mEditAim->setText(QString("%1").arg(mAimPoint,0,'f',0));
	ui.mEditWind->setText(QString("%1").arg(mXWind,0,'f',0));
	ui.mEditTemp->setText(QString("%1").arg(mTemp,0,'f',0));
}
// --------------------------------------------------------
void
External_Ballistics::showLine(QString line)
{
	ui.mShowResults->append(line);
}
// --------------------------------------------------------
void
External_Ballistics::showOutput()
{
	int yards = 0;

	// Begin with data every yard
	while((yards < mVelocity.count()) && (yards * data_interval) < 50) {
		doOutputLine(yards);
		++yards;
	}

	// At 50-100 yards, go up by 2
	while((yards < mVelocity.count()) && (yards * data_interval) < 100) {
		doOutputLine(yards);
		yards += 2;
	}

	// Now go up by 5 yard intervals
	while((yards < mVelocity.count()) && (yards * data_interval) < 400) {
		doOutputLine(yards);
		yards += 5;
	}

}
// --------------------------------------------------------


